import unittest
import numpy as np
from scipy.stats import bootstrap as sboot
from foostrap import foostrap

drng = np.random.default_rng
rng = drng(0)
rand_arr1d = np.concatenate([np.zeros(500), rng.normal(size= 500)])
rand_arr2d = np.concatenate([rand_arr1d.reshape(-1,1), rng.lognormal(size= 1000).reshape(-1,1)], axis= 1)
rand_arr2d_corr = rand_arr2d.copy()
rand_arr2d_corr[:,1] = 2*rand_arr1d + 5 + rng.normal(size=1000)

print('Test array shapes', rand_arr1d.shape, rand_arr2d.shape, rand_arr2d_corr.shape)

class TestNoInput(unittest.TestCase):
    def test_empty1(self):
        with self.assertRaises(ValueError):
            foostrap(np.array([]))

    def test_empty2(self):
        with self.assertRaises(ValueError):
            foostrap(rand_arr1d, x2= np.array([]))

    def test_all_invalid1(self):
        with self.assertRaises(ValueError):
            foostrap(np.array([np.nan, np.inf]))

    def test_all_invalid2(self):
        with self.assertRaises(ValueError):
            foostrap(rand_arr2d, x2= np.array([[np.nan, 1], [1, np.inf]]))

class TestDimensions(unittest.TestCase):
    def test_mismatch(self):
        with self.assertRaises(ValueError):
            foostrap(rand_arr1d, rand_arr2d)

    def test_3d(self):
        with self.assertRaises(ValueError):
            foostrap(np.ones((3, 3, 3)))

class TestLowVariaion(unittest.TestCase):
    def test_all_equal1(self):
        foostrap(np.full(3, 0.5))

    def test_all_equal2(self):
        foostrap(rand_arr1d, np.full(3, 0.5))

    def test_single_value1(self):
        foostrap(np.array([1]))

    def test_single_value2(self):
        foostrap(np.array([1]), np.array([1]))

    def test_single_value_ratio1(self):
        foostrap(np.array([1,1]).reshape((1,2)))

    def test_single_value_ratio2(self):
        foostrap(np.array([1,1]).reshape((1,2)), np.array([1,1]).reshape((1,2)))

    def test_low_nboot(self):
        foostrap(rand_arr1d, rand_arr1d, boot_samples= 1)

class TestDtypes(unittest.TestCase):
    def test_floats(self):
        foostrap(rand_arr1d.astype(np.float64), rand_arr1d.astype(np.float32))

    def test_ints(self):
        foostrap(np.array([1,2,3]).astype(np.int64), np.array([1,2,3]).astype(np.int32))

    def test_mix_float_ints(self):
        foostrap(np.array([1,2,3]).astype(np.int64), np.array([1,2,3]).astype(np.float32))

class TestIllegalMath(unittest.TestCase):
    def test_zeros_denom(self):
        with self.assertWarns(UserWarning):
            foostrap(np.concatenate((np.ones((100,2)), np.array([1,0]).reshape(-1,2))), random_state=0, parallel=False)

class TestMixedSparsity(unittest.TestCase):
    def test_dense_sparse(self):
        res = foostrap(np.full(3, 2), rand_arr1d, boot_samples= 1000)
        self.assertFalse(np.isnan(res.ci).any() or np.isnan(res.boot_samples).any(), 'There are nan values in output')

    def test_bin_sparse(self):
        res = foostrap(rng.binomial(1,0.5,200), rand_arr1d, boot_samples= 1000)
        self.assertFalse(np.isnan(res.ci).any() or np.isnan(res.boot_samples).any(), 'There are nan values in output')

    def test_bin_dense(self):
        res = foostrap(rng.binomial(1,0.5,50), np.full(3, 2), boot_samples= 1000)
        self.assertFalse(np.isnan(res.ci).any() or np.isnan(res.boot_samples).any(), 'There are nan values in output')

#Statistic functions for the Scipy execution
def mean_dif(x, y, axis= -1):
    return np.mean(x, axis= axis) - np.mean(y, axis= axis)
def std_dif(x, y, axis= -1):
    return np.std(x, axis= axis) - np.std(y, axis= axis)
def quant(x, axis= -1):
    return np.quantile(x, 0.6, axis= axis)
def quant_dif(x, y, axis= -1):
    return np.quantile(x, 0.6, axis= axis) - np.quantile(y, 0.6, axis= axis)
def ratio_dif(x, y, axis= -1):
    return x[0].sum(axis= axis) / x[1].sum(axis= axis) - y[0].sum(axis= axis) / y[1].sum(axis= axis)
def ratio(x, axis= -1):
    return x[0].sum(axis= axis) / x[1].sum(axis= axis)
def wmean(x, axis= -1):
    return x.prod(axis= 0, keepdims= True).sum(axis= axis).squeeze() / x[1].sum(axis= axis)
def wmean_dif(x, y, axis= -1):
    return x.prod(axis= 0, keepdims= True).sum(axis= axis).squeeze() / x[1].sum(axis= axis) - y.prod(axis= 0, keepdims= True).sum(axis= axis).squeeze() / y[1].sum(axis= axis)
def pearson(x, axis= -1):
    if x.ndim == 3:
        return [np.corrcoef(x[0,k,:],x[1,k,:])[0,1] for k in range(x.shape[1])]
    else:
        return np.corrcoef(x[0,:],x[1,:])[0,1]
def pearson_dif(x, y, axis= -1):
    if x.ndim == 3:
        return [np.corrcoef(x[0,k,:],x[1,k,:])[0,1] - np.corrcoef(y[0,k,:],y[1,k,:])[0,1] for k in range(x.shape[1])]
    else:
        return np.corrcoef(x[0,:],x[1,:])[0,1] - np.corrcoef(y[0,:],y[1,:])[0,1]

#Comparison of CIs against Scipy
class TestScipyClose(unittest.TestCase):
    def test_2sample_dense_2d(self):
        rfoo = foostrap(rand_arr2d, rand_arr2d+0.1, random_state=drng(1), parallel=False, boot_samples= 30000)
        rsci = sboot((rand_arr2d, rand_arr2d+0.1), ratio_dif, vectorized= True, random_state= drng(1), batch= 1000, n_resamples= 30000)
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_1sample_dense_2d(self):
        rfoo = foostrap(rand_arr2d, random_state=drng(1), parallel=False)
        rsci = sboot((rand_arr2d,), ratio, vectorized= True, random_state= drng(1), batch= 1000)
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')
        
    def test_2sample_sparse_less_conflvl(self):
        rfoo = foostrap(rand_arr1d, rand_arr1d*1.1, random_state=drng(1), parallel=False, conf_lvl= 0.9, alternative= 'less')
        rsci = sboot((rand_arr1d, rand_arr1d*1.1), mean_dif, vectorized= True, random_state= drng(1), batch= 1000, n_resamples= 10000, alternative= 'less', confidence_level= 0.9)
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_2sample_binary_greater_conflvl_cimethod(self):
        rfoo = foostrap(rand_arr1d!=0, rand_arr1d!=0, random_state=drng(1), parallel=False, conf_lvl= 0.9, alternative= 'greater', ci_method='percentile')
        rsci = sboot((1*(rand_arr1d!=0), 1*(rand_arr1d!=0)), mean_dif, vectorized= True, random_state= drng(1), batch= 1000, alternative= 'greater', confidence_level= 0.9, method= 'percentile')
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_1sample_dense_less_conflvl(self):
        rfoo = foostrap(rand_arr1d, random_state=drng(1), parallel=False, conf_lvl= 0.85, alternative= 'less', ignore_sparse_below=1.0)
        rsci = sboot((rand_arr1d,), np.mean, vectorized= True, random_state= drng(1), batch= 1000, n_resamples= 10000, alternative= 'less', confidence_level= 0.85)
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_1sample_sparse_greater_cimethod(self):
        rfoo = foostrap(rand_arr1d, random_state=drng(1), parallel=False, alternative= 'greater', ci_method='percentile')
        rsci = sboot((rand_arr1d,), np.mean, vectorized= True, random_state= drng(1), batch= 1000, n_resamples= 10000, alternative= 'greater', method= 'percentile')
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_1sample_binary_conflvl(self):
        rfoo = foostrap(rand_arr1d!=0, random_state=drng(1), parallel=False, conf_lvl= 0.98)
        rsci = sboot((1*(rand_arr1d!=0),), np.mean, vectorized= True, random_state= drng(1), batch= 1000, n_resamples= 10000, confidence_level= 0.98)
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_2sample_dense_std(self):
        rfoo = foostrap(rand_arr1d, rand_arr1d*1.1, statistic= 'std', random_state=drng(1), parallel=False, ignore_sparse_below=1.0)
        rsci = sboot((rand_arr1d, rand_arr1d*1.1), std_dif, vectorized= True, random_state= drng(1), batch= 1000)
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_2sample_sparse_std(self):
        rfoo = foostrap(rand_arr1d, rand_arr1d*1.1, statistic= 'std', random_state=drng(1), parallel=False)
        rsci = sboot((rand_arr1d, rand_arr1d*1.1), std_dif, vectorized= True, random_state= drng(1), batch= 1000)
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_1sample_binary_std(self):
        rfoo = foostrap(rand_arr1d!=0, statistic= 'std', random_state=drng(1), parallel=False)
        rsci = sboot((1*(rand_arr1d!=0),), np.std, vectorized= True, random_state= drng(1), batch= 1000)
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_2sample_dense_quant(self):
        rfoo = foostrap(rand_arr1d, rand_arr1d*1.1, statistic= 'quantile', q= 0.6, random_state=drng(1), parallel=False, ignore_sparse_below=1.0, ci_method= 'percentile')
        rsci = sboot((rand_arr1d, rand_arr1d*1.1), quant_dif, vectorized= True, random_state= drng(1), batch= 1000, method= 'percentile')
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_2sample_sparse_quant(self):
        rfoo = foostrap(rand_arr1d, rand_arr1d*1.1, statistic= 'quantile', q= 0.6, random_state=drng(1), parallel=False, ci_method= 'percentile')
        rsci = sboot((rand_arr1d, rand_arr1d*1.1), quant_dif, vectorized= True, random_state= drng(1), batch= 1000, method= 'percentile')
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_2sample_binary_quant(self):
        rfoo = foostrap(rand_arr1d!=0, rand_arr1d>0.5, statistic= 'quantile', q= 0.6, random_state=drng(1), parallel=False, ci_method= 'percentile')
        rsci = sboot((1*(rand_arr1d!=0), 1*(rand_arr1d>0.5)), quant_dif, vectorized= True, random_state= drng(1), batch= 1000, method= 'percentile')
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_2sample_dense_2d_wmean(self):
        rfoo = foostrap(rand_arr2d, rand_arr2d+0.1, statistic= 'wmean', random_state=drng(1), parallel=False)
        rsci = sboot((rand_arr2d, rand_arr2d+0.1), wmean_dif, vectorized= True, random_state= drng(1), batch= 1000)
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_1sample_dense_2d_wmean(self):
        rfoo = foostrap(rand_arr2d, statistic= 'wmean', random_state=drng(1), parallel=False)
        rsci = sboot((rand_arr2d,), wmean, vectorized= True, random_state= drng(1), batch= 1000)
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_2sample_dense_2d_pearson(self):
        rfoo = foostrap(rand_arr2d_corr, rand_arr2d_corr+rand_arr2d, statistic= 'pearson', random_state=drng(1), parallel=False)
        rsci = sboot((rand_arr2d_corr, rand_arr2d_corr+rand_arr2d), pearson_dif, vectorized= True, random_state= drng(1), batch= 1000)
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')

    def test_1sample_dense_2d_pearson(self):
        rfoo = foostrap(rand_arr2d_corr, statistic= 'pearson', random_state=drng(1), parallel=False)
        rsci = sboot((rand_arr2d_corr,), pearson, vectorized= True, random_state= drng(1), batch= 1000)
        self.assertAlmostEqual(rfoo.ci[0], rsci.confidence_interval.low, places=2)
        self.assertAlmostEqual(rfoo.ci[1], rsci.confidence_interval.high, places=2)
        self.assertFalse(np.isnan(rfoo.ci).any() or np.isnan(rfoo.boot_samples).any(), 'There are nan values in output')
        
if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
import numpy as np
from numba import njit, objmode, prange

@njit(parallel= True, cache=True)
def core_parallelizer(sampler_func1, sampler_func2, stat_func1, stat_func2, x1, x2, gens, n1, n2, n_boot, arg):
    # Use numba prange to distribute bootstrap sampling iterations equally among threads. Each thread has its own generator
    nt = len(gens)
    thread_samps = np.array([n_boot // nt] * nt) + np.array([1] * (n_boot % nt) + [0] * (nt - (n_boot % nt)))
    idxs = [0] + list(np.cumsum(thread_samps))
    boot_stat = np.empty(n_boot)
    for i in prange(nt):
        local_res = sampler_func1(x1, thread_samps[i], gens[i], n1, stat_func1, arg)
        if n2 > 0:
            local_res -= sampler_func2(x2, thread_samps[i], gens[i], n2, stat_func2, arg)
        boot_stat[idxs[i] : idxs[i+1]] = local_res
    return boot_stat

@njit(fastmath=True, cache=True)
def bstrap_sampler_1d(x, n_boot, gen, n, stat_func, arg):
    # Generate n_boot bootstrap samples for 1D x using generator gen and compute the statistic stat_func
    boot_stat = np.empty(n_boot)
    for i in range(n_boot):
        boot_stat[i] = stat_func(x[gen.integers(0, n, size= n, dtype= np.uint64)])
    return boot_stat

@njit(fastmath=True, cache=True)
def bstrap_sampler_1d_sparse(x, n_boot, gen, n, stat_func, arg):
    # Generate n_boot bootstrap samples for 1D sparse x using generator gen and compute the statistic stat_func
    nnz = len(x)
    p = nnz / n
    with objmode(nzboot='int64[:]'):
        nzboot = gen.binomial(n, p, n_boot)
    boot_stat = np.empty(n_boot)
    for i in range(n_boot):
        boot_stat[i] = stat_func(x[gen.integers(0, nnz, size= nzboot[i], dtype= np.uint64)], n)
    return boot_stat

@njit(fastmath=True, cache=True)
def bstrap_sampler_1d_bin(x, n_boot, gen, n, stat_func, arg):
    # Generate n_boot bootstrap samples for 1D binary x using generator gen and compute the statistic stat_func
    p = x.sum() / n
    with objmode(n_ones='int64[:]'):
        n_ones = gen.binomial(n, p, n_boot)
    return stat_func(n_ones, n)

@njit(fastmath=True, cache=True)
def bstrap_sampler_1d_arg(x, n_boot, gen, n, stat_func, arg):
    # Generate n_boot bootstrap samples for 1D x using generator gen and compute the statistic stat_func, accepting an argument arg
    boot_stat = np.empty(n_boot)
    for i in range(n_boot):
        boot_stat[i] = stat_func(x[gen.integers(0, n, size= n, dtype= np.uint64)], arg)
    return boot_stat

@njit(fastmath=True, cache=True)
def bstrap_sampler_1d_sparse_arg(x, n_boot, gen, n, stat_func, arg):
    # Generate n_boot bootstrap samples for 1D sparse x using generator gen and compute the statistic stat_func
    nnz = len(x)
    p = nnz / n
    with objmode(nzboot='int64[:]'):
        nzboot = gen.binomial(n, p, n_boot)
    boot_stat = np.empty(n_boot)
    for i in range(n_boot):
        boot_stat[i] = stat_func(x[gen.integers(0, nnz, size= nzboot[i], dtype= np.uint64)], n, arg)
    return boot_stat

@njit(fastmath=True, cache=True)
def bstrap_sampler_1d_bin_arg(x, n_boot, gen, n, stat_func, arg):
    # Generate n_boot bootstrap samples for 1D binary x using generator gen and compute the statistic stat_func
    p = x.sum() / n
    with objmode(n_ones='int64[:]'):
        n_ones = gen.binomial(n, p, n_boot)
    return stat_func(n_ones, n, arg)

@njit(fastmath=True, cache=True)
def bstrap_sampler_2d(x, n_boot, gen, n, stat_func, arg):
    # Generate n_boot bootstrap samples for 2D x using generator gen and compute the statistic stat_func
    boot_stat = np.empty(n_boot)
    for i in range(n_boot):
        boot_stat[i] = stat_func(x[:,gen.integers(0, n, size= n, dtype= np.uint64)])
    return boot_stat

sampler_map = {(False,False,False) : bstrap_sampler_1d,
               (False,True,False) : bstrap_sampler_1d_sparse,
               (True,False,False) : bstrap_sampler_1d_bin,
               (False,False,True) : bstrap_sampler_1d_arg,
               (False,True,True) : bstrap_sampler_1d_sparse_arg,
               (True,False,True) : bstrap_sampler_1d_bin_arg,
               '2d' : bstrap_sampler_2d,
              }


# -*- coding: utf-8 -*-
import numpy as np
from scipy.stats import mode
from .estimators import stat_func_map, jack_func_map, ci_est_percent, ci_est_bca
from .samplers import sampler_map
import warnings

def check_sparsity(x, n, min_zeros= 0.1):
    # Flag data if binary. If not binary, but sparse, remove the zeros from the array
    xz = x == 0
    is_bin = np.all(xz | (x == 1))
    if xz.sum() / n > min_zeros and n > 100:
        ret = x[~xz].copy()
    else:
        ret = x
    return is_bin, ret

def warn_pearson(x, n, nboot, threshold):
    # Warn user if probability of a sample of equal values exceeds threshold
    m,k = mode(x)
    prob = 1.0 - (1.0-(k/n)**n)**nboot
    if prob > threshold:
        warnings.warn(f'The value {m} is repeated {k} times out of {n} observations. There is a {np.round(prob, int(np.ceil(-np.log10(threshold))))*100} % probability of getting at least 1 nan sample out of {nboot} due to zero denominator.', stacklevel=2)

def preproc_args(x1, x2, statistic, q, boot_samples, conf_lvl, alternative, ci_method, random_state, parallel, ignore_sparse_below):
    # Check validity of each argument, and preprocess them
    
    # x1
    x1t = np.asarray(x1).squeeze()
    if x1t.ndim > 2:
        raise ValueError("x1 must be at most 2-dimensional after squeezing.")
    elif x1t.ndim == 2 and x1t.shape[1] != 2:
        raise ValueError("If x1 is 2-dimensional, the second dimension must be of length 2.")
    elif x1t.ndim == 0:
        x1t = x1t.reshape(-1)
    is_1d = x1t.ndim == 1
    x1_valid = np.isfinite(x1t) if is_1d else np.isfinite(x1t).all(axis=1)
    n1 = x1_valid.sum()
    if n1 == 0:
        if len(x1t) == 0:
            raise ValueError("x1 cannot be empty")
        else:
            raise ValueError("x1 must contain at least 1 valid number")
    x1t = x1t[x1_valid] if is_1d else x1t[x1_valid,:].T.copy()
    if x1t.dtype == bool:
        x1t = x1t.astype(np.uint8)
    
    # x2
    if x2 is not None:
        x2t = np.asarray(x2).squeeze()
        if x2t.ndim == 0:
            x2t = x2t.reshape(-1)
        elif x2t.ndim != x1t.ndim:
            raise ValueError("x2 must have the same number of dimensions as x1.")
        elif x2t.ndim == 2 and x2t.shape[1] != 2:
            raise ValueError("If x2 is 2-dimensional, the second dimension must be of length 2.")
        x2_valid = np.isfinite(x2t) if is_1d else np.isfinite(x2t).all(axis=1)
        n2 = x2_valid.sum()
        if n2 == 0:
            raise ValueError("x2 must have at least 1 valid number, or be omitted.")
        else:
            x2t = x2t[x2_valid] if is_1d else x2t[x2_valid,:].T.copy()
    else:
        n2 = 0
        x2t = np.array([], dtype= x1t.dtype) if is_1d else np.array([], dtype= x1t.dtype).reshape((2,-1))
    if x2t.dtype == bool:
        x2t = x2t.astype(np.uint8)
        
    # statistic
    if statistic not in ('auto','mean','std','quantile','ratio','wmean','pearson'):
        raise ValueError("statistic must be one of ('auto','mean','std','quantile','ratio','wmean','pearson')")
    if statistic in ('mean','std','quantile') and not is_1d:
        raise ValueError(f"statistic {statistic} is only available for 1-dimensional arrays")
    if statistic in ('ratio','wmean','pearson') and is_1d:
        raise ValueError(f"statistic {statistic} is only available for 2-dimensional arrays")
    if statistic == 'auto':
        statistic = 'mean' if is_1d else 'ratio'

    # q
    if statistic == 'quantile' and not (isinstance(q, (float,int)) and 0 <= q <= 1):
        raise ValueError("q must be a single number between 0 and 1")
    
    # boot_samples
    if not isinstance(boot_samples, int) or boot_samples <= 0:
        raise ValueError("boot_samples must be an integer greater than 0.")
    
    # conf_lvl
    if not (isinstance(conf_lvl, float) and 0 < conf_lvl < 1):
        raise ValueError("conf_lvl must be a float greater than 0 and less than 1.")

    # alternative
    if alternative not in ('two-sided','less','greater'):
        raise ValueError("alternative must be either 'two-sided', 'less' or 'greater'")
    if alternative == 'two-sided':
        ci_alphas = (0.5 - conf_lvl/2.0, 0.5 + conf_lvl/2.0)
    elif alternative == 'less':
        ci_alphas = (-np.inf, conf_lvl)
    else:
        ci_alphas = (1.0 - conf_lvl, np.inf)
    
    # ci_method
    if ci_method not in ['BCa', 'percentile']:
        raise ValueError("ci_method must be either 'BCa' or 'percentile'.")
    
    # random_state
    if not (random_state is None or isinstance(random_state, (int, np.random.Generator, np.random.SeedSequence))):
        raise ValueError("random_state must be None, an integer, numpy.random.Generator, or numpy.random.SeedSequence.")
    if random_state is None:
        rng = np.random.Generator(np.random.SFC64())
    elif isinstance(random_state, (int, np.random.SeedSequence)):
        rng = np.random.Generator(np.random.SFC64(random_state))
    else:
        rng = random_state
        
    # parallel
    if not isinstance(parallel, bool):
        raise ValueError("parallel must be a boolean.")
    
    # ignore_sparse_below
    if not (isinstance(ignore_sparse_below, (float, int)) and 0 <= ignore_sparse_below <= 1):
        raise ValueError("ignore_sparse_below must be a float greater or equal to 0 and less or equal to 1.")

    # If arrays are 1D, check sparsity of samples
    if is_1d:
        is_bin1, x1t = check_sparsity(x1t, n1, min_zeros= ignore_sparse_below)
        if n2 > 0:
            is_bin2, x2t = check_sparsity(x2t, n2, min_zeros= ignore_sparse_below)
        else:
            is_bin2 = True
        # Pick corresponding sampler and statistic function
        is_sparse1 = (not is_bin1) and (len(x1t) < n1)
        is_sparse2 = (not is_bin2) and (len(x2t) < n2)
        extra_arg = statistic in ('quantile',)
        sampler_func1 = sampler_map[(is_bin1, is_sparse1, extra_arg)]
        sampler_func2 = sampler_map[(is_bin2, is_sparse2, extra_arg)]
        stat_func1 = stat_func_map[(statistic, is_bin1, is_sparse1)]
        stat_func2 = stat_func_map[(statistic, is_bin2, is_sparse2)]
        if is_bin1 and is_bin2:
            parallel = False
    else:
        # If using paired statistics, sparsity doesn't apply
        sampler_func1, sampler_func2 = sampler_map['2d'], sampler_map['2d']
        stat_func1, stat_func2 = stat_func_map[statistic], stat_func_map[statistic]
    
    # Get jacknife estimator
    jack_func = jack_func_map[statistic] if statistic != 'quantile' else lambda x, n1, n2 : jack_func_map['quantile'](x, n1, n2, q)
    
    # Set function for confidence interval estimation (percentile or BCa)
    if n1 == 1 or n2 == 1:
        ci_method = 'percentile'
    ci_func = ci_est_percent if ci_method == 'percentile' else ci_est_bca

    # Warn user if doing dividing by sums that contain zeros
    if statistic in ('ratio','wmean') and (np.any(x1t[1,:] == 0) or np.any(x2t[1,:] == 0)):
        warnings.warn('Zeros detected in the denominator values. This could cause a division by zero over a random resampling. These samples will be recorded as nan', stacklevel=2)

    # Warn user if many obsevations for the pearson correlation are equal (risk of zero variance in denominator)
    if statistic in ('pearson',):
        warn_pearson(x1t[0,:], n1, boot_samples, 0.001)
        warn_pearson(x1t[1,:], n1, boot_samples, 0.001)
        if n2 > 0:
            warn_pearson(x2t[0,:], n2, boot_samples, 0.001)
            warn_pearson(x2t[1,:], n2, boot_samples, 0.001)
    
    return x1t, x2t, rng, parallel, n1, n2, ci_alphas, sampler_func1, sampler_func2, stat_func1, stat_func2, jack_func, ci_func



# -*- coding: utf-8 -*-
import numpy as np
from numba import get_num_threads
from dataclasses import dataclass
from .preproc import preproc_args
from .samplers import core_parallelizer

@dataclass
class BootRes:
    # Bootstrap results. Confidence interval tuple and array of samples
    ci: tuple
    boot_samples: np.array

def foostrap(x1, x2= None, statistic= 'auto', q= 0.5, boot_samples= 10000, conf_lvl= 0.95, alternative= 'two-sided', ci_method= 'BCa', random_state= None, parallel= True, ignore_sparse_below= 0.1):
    '''
    Perform parallel bootstrap sampling and confidence interval estimation.
    The available statistics are:
    - For 1-dimensional data: mean, standard deviation, quantile q
    - For 2-dimensional paired data: ratio of sums, weighted mean and pearson correlation
    To compare two independent samples (e.g. mean difference), use the argument x2 for the second sample.

    Parameters:
    - x1 (numpy.ndarray): Primary sample array. If observations are paired, the shape must be a 2-column array.
    - x2 (numpy.ndarray, optional): Second sample to compare against x1, as statistic(x1) - statistic(x2). Default is None.
    - statistic (one of 'mean','std','quantile','ratio','wmean','pearson','auto'): the statistic to compute over each sample. Default 'auto' ('mean' for 1D sample, 'ratio' for 2D sample).
    - q (float): probability for the 'quantile' statistic. Ignored otherwise. Default 0.5 (the median)
    - boot_samples (int): Number of bootstrap samples to generate. Must be greater than zero. Default is 10000.
    - conf_lvl (float): Confidence level for the confidence interval, between 0 and 1. Default is 0.95.
    - alternative (str): Type of confidence interval. 'two-sided': with upper and lower bound, 'less': only upper bound, 'greater': only lower bound. Default 'two-sided'.
    - ci_method (str): Method for confidence interval estimation. Options are 'BCa' for Bias-Corrected and Accelerated or 'percentile'. Default is 'BCa'.
    - random_state (None, int, numpy.random.Generator, or numpy.random.SeedSequence): Seed or numpy random generator for reproducibility. Default is None.
    - parallel (bool): Whether to use parallel processing for bootstrap sampling. Default is True.
    - ignore_sparse_below (float): Threshold under which sparse data is treated as dense, to avoid the overhead of a separate sampling. Between 0 and 1. Default is 0.1.
    
    Returns:
    BootRes: A dataclass containing the confidence interval (ci) as a tuple and the bootstrap samples (boot_samples) as a numpy array.
    
    Example 1: mean statistic, 1-sample
    >>> x1 = np.random.normal(0, 1, size=100)
    >>> result = foostrap(x1)
    >>> print(result.ci)
    (-0.199, 0.204)

    Example 2: weighted mean difference statistic, 2-sample
    >>> x1 = np.random.lognormal(0, 1, size= (100, 2))
    >>> x2 = np.random.lognormal(0, 1, size= (100, 2))
    >>> result = foostrap(x1, x2, statistic= 'wmean')
    >>> print(result.ci)
    (-1.04, 0.194)
    '''
    # Validate and preprocess arguments
    x1t, x2t, rng, parallel, n1, n2, ci_alphas, sampler_func1, sampler_func2, stat_func1, stat_func2, jack_func, ci_func = \
        preproc_args(x1, x2, statistic, q, boot_samples, conf_lvl, alternative, ci_method, random_state, parallel, ignore_sparse_below)

    # If parallel, split generators and call the numba multicore wrapper
    if parallel:
        nt = get_num_threads()
        gens = tuple(rng.spawn(nt))
        boot_stat = core_parallelizer(sampler_func1, sampler_func2, stat_func1, stat_func2, x1t, x2t, gens, n1, n2, boot_samples, q)
    else:
        # Single core directly calls the sampler
        boot_stat = sampler_func1(x1t, boot_samples, rng, n1, stat_func1, q)
        if n2 > 0:
            boot_stat -= sampler_func2(x2t, boot_samples, rng, n2, stat_func2, q)

    # Get confidence intervals for the bootstrap samples
    quants = ci_func(boot_stat, ci_alphas, jack_func, x1t, x2t, n1, n2)
    return BootRes(ci= tuple(quants), boot_samples= boot_stat)



# -*- coding: utf-8 -*-
import numpy as np
from numba import njit
from scipy.stats import norm

@njit(fastmath=True, cache=True)
def numba_mean(a):
    # Compute the mean of a
    n = len(a)
    s = a.dtype.type(0)
    for i in range(n):
        s += a[i]
    return s / n

@njit(fastmath=True, cache=True)
def numba_mean_sparse(a, n):
    # Compute the mean of sparse a
    nnz = len(a)
    s = a.dtype.type(0)
    for i in range(nnz):
        s += a[i]
    return s / n

@njit(fastmath=True, cache=True)
def numba_mean_bin(a, n):
    # Compute the mean of binary a
    return a / n

@njit(fastmath=True, cache=True)
def numba_std(a):
    # Compute the standard deviation of a
    n = len(a)
    sx = a.dtype.type(0)
    sx2 = sx
    for i in range(n):
        sx += a[i]
        sx2 += a[i] ** 2
    return np.sqrt(sx2 / n - (sx / n) ** 2)

@njit(fastmath=True, cache=True)
def numba_std_sparse(a, n):
    # Compute the standard deviation of sparse a
    nnz = len(a)
    sx = a.dtype.type(0)
    sx2 = sx
    for i in range(nnz):
        sx += a[i]
        sx2 += a[i] ** 2
    return np.sqrt(sx2 / n - (sx / n) ** 2)

@njit(fastmath=True, cache=True)
def numba_std_bin(a, n):
    # Compute the standard deviation of binary a
    p = a / n
    return np.sqrt(p * (1.0 - p))

@njit(fastmath=True, cache=True)
def quantile_dense(a, q):
    # Compute the q-quantile of dense a
    return np.quantile(a, q)

@njit(fastmath=True, cache=True)
def quantile_sparse(a, n, q):
    # Compute the q-quantile of sparse a
    return np.quantile(np.concatenate((np.zeros(n-len(a), dtype= a.dtype), a)), q)

@njit(fastmath=True, cache=True)
def quantile_bin(a, n, q):
    # Compute the q-quantile of binary a
    nz = n - a
    idx = q * (n-1) + 1.0
    idx_lo = int(np.floor(idx))
    return 1.0 * (nz < idx_lo) + (idx - idx_lo) * (nz == idx_lo)

@njit(fastmath= True, cache=True)
def ratio_rows(a):
    # Compute the ratio of sums of consecutive rows
    n = a.shape[1]
    sn = a.dtype.type(0)
    sd = sn
    for i in range(n):
        sn += a[0,i]
        sd += a[1,i]
    if sd != 0:
        return sn / sd
    else:
        return np.nan

@njit(fastmath= True, cache=True)
def wmean_rows(a):
    # Compute the weighted mean of consecutive rows
    n = a.shape[1]
    sxw = a.dtype.type(0)
    sw = sxw
    for i in range(n):
        sxw += a[0,i] * a[1,i]
        sw += a[1,i]
    if sw != 0:
        return sxw / sw
    else:
        return np.nan

@njit(fastmath= True, cache=True)
def pearson_rows(a):
    # Compute the pearson correlation of consecutive rows
    n = a.shape[1]
    sx = a.dtype.type(0)
    sy, sxy, sx2, sy2 = sx, sx, sx, sx
    for i in range(n):
        sx += a[0,i]
        sx2 += a[0,i] ** 2
        sy += a[1,i]
        sy2 += a[1,i] ** 2
        sxy += a[0,i] * a[1,i]
    denom = np.sqrt((n * sx2 - sx ** 2)*(n * sy2 - sy ** 2))
    if denom != 0.0:
        return (n * sxy - sx * sy) / denom
    else:
        return np.nan

def jacknife_mean(x, n1, n2):
    # Generate jacknife estimates for the mean
    s = x.sum()
    return s / n1, \
        (s - np.pad(x, (0, n1 - len(x) + n2))) / \
        np.concatenate((np.full(n1,n1-1,dtype=np.float64), \
                        np.full(n2,n1,dtype=np.float64)))

def jacknife_std(x, n1, n2):
    # Generate jacknife estimates for the standard deviation
    s = x.sum()
    s2 = (x**2).sum()
    denom = np.concatenate((np.full(n1,n1-1,dtype=np.float64), np.full(n2,n1,dtype=np.float64)))
    x_ext = np.pad(x, (0, n1 - len(x) + n2))
    return np.sqrt(s2 / n1 - (s / n1)**2), \
        np.sqrt((s2 - x_ext**2)/denom - ((s - x_ext)/denom)**2)

def jacknife_quantile(x, n1, n2, q):
    # Generate jacknife estimates for the q-quantile
    xs = np.sort(np.pad(x, (n1 - len(x), 0)))
    stat = np.quantile(xs, q)
    idx = q * (n1-2)
    idx_lo, idx_hi = np.full(n1,int(np.floor(idx))), np.full(n1,int(np.ceil(idx)))
    idx_lo[:idx_lo[0]+1] += 1
    idx_hi[:idx_hi[0]+1] += 1
    xlo, xhi = xs[idx_lo], xs[idx_hi]
    return stat, np.pad((xhi - xlo) * (idx - idx_lo), (0, n2), constant_values= stat)

def jacknife_ratio(x, n1, n2):
    # Generate jacknife estimates for the ratio of sums
    s = x.sum(axis= 1)
    jacks = s.reshape(2,1) - np.pad(x, ((0,0),(0,n2)))
    return s[0] / s[1], jacks[0,:] / jacks[1,:]

def jacknife_wmean(x, n1, n2):
    # Generate jacknife estimates for the weighted mean
    xw = np.dot(x[0,:], x[1,:])
    w = x[1,:].sum()
    xw_ext = np.pad(x, ((0,0), (0,n2)))
    return xw / w, (xw - xw_ext[0,:] * xw_ext[1,:]) / (w - xw_ext[1:])

def jacknife_pearson(x, n1, n2):
    # Generate jacknife estimates for the pearson correlation
    sx, sy = x.sum(axis= 1)
    sx2, sy2 = (x ** 2).sum(axis= 1)
    sxy = np.dot(x[0,:], x[1,:])
    stat = (n1 * sxy - sx * sy) / np.sqrt((n1 * sx2 - sx ** 2)*(n1 * sy2 - sy ** 2))
    xy_ext = np.pad(x, ((0,0), (0,n2)))
    n_ext = np.concatenate((np.full(n1,n1-1,dtype=np.float64), np.full(n2,n1,dtype=np.float64)))
    jacks = (n_ext * (sxy - xy_ext[0,:] * xy_ext[1,:]) - (sx - xy_ext[0,:]) * (sy - xy_ext[1,:])) / \
        np.sqrt((n_ext * (sx2 - xy_ext[0,:] ** 2) - (sx - xy_ext[0,:])**2) * (n_ext * (sy2 - xy_ext[1,:] ** 2) - (sy - xy_ext[1,:])**2))
    return stat, jacks
    
def jacknife_wrapper(x1, x2, n1, n2, jack_func):
    # Call the jack_func statistic over 1 or 2 samples and compute the difference
    samp_stat, jacks = jack_func(x1, n1, n2)
    if n2 > 0:
        samp_stat2, jacks2 = jack_func(x2, n2, n1)
        samp_stat -= samp_stat2
        jacks -= np.flip(jacks2)
    return samp_stat, jacks

def ci_est_percent(samps, ci_alphas, jack_func, x1, x2, n1, n2):
    # Compute percentiles directly from bootstrap samples
    if ci_alphas[0] == -np.inf:
        return (-np.inf, np.nanquantile(samps, ci_alphas[1]))
    elif ci_alphas[1] == np.inf:
        return (np.nanquantile(samps, ci_alphas[0]), np.inf)
    else:
        return np.nanquantile(samps, ci_alphas)

def ci_est_bca(samps, ci_alphas, jack_func, x1, x2, n1, n2):
    # Get jacknife estimates and sample statistic
    samp_stat, jacks = jacknife_wrapper(x1, x2, n1, n2, jack_func)
    # Compute the bias term
    b = norm.ppf((1.0*(samps < samp_stat) + 0.5*(samps == samp_stat)).mean())
    # Compute the acceleration term
    jack_mean = jacks.mean()
    acc = ((jack_mean - jacks)**3).sum() / 6 / (((jack_mean - jacks)**2).sum())**1.5
    # Final formula for corrected quantiles
    ci_lo, ci_hi = -np.inf, np.inf
    if ci_alphas[0] != -np.inf:
        per_norm_lo = b + norm.ppf(ci_alphas[0])
        ci_lo = max(0.0, norm.cdf(b + per_norm_lo/(1.0 - acc*per_norm_lo)))
    if ci_alphas[1] != np.inf:
        per_norm_hi = b + norm.ppf(ci_alphas[1])
        ci_hi = min(1.0, norm.cdf(b + per_norm_hi/(1.0 - acc*per_norm_hi)))
    return ci_est_percent(samps, (ci_lo, ci_hi), jack_func, x1, x2, n1, n2)

stat_func_map = {('mean',False,False) : numba_mean,
                 ('mean',False,True) : numba_mean_sparse,
                 ('mean',True,False) : numba_mean_bin,
                 ('std',False,False) : numba_std,
                 ('std',False,True) : numba_std_sparse,
                 ('std',True,False) : numba_std_bin,
                 ('quantile',False,False) : quantile_dense,
                 ('quantile',False,True) : quantile_sparse,
                 ('quantile',True,False) : quantile_bin,
                 'ratio' : ratio_rows,
                 'wmean' : wmean_rows,
                 'pearson' : pearson_rows,
                }

jack_func_map = {'mean' : jacknife_mean,
                 'std' : jacknife_std,
                 'quantile' : jacknife_quantile,
                 'ratio' : jacknife_ratio,
                 'wmean' : jacknife_wmean,
                 'pearson' : jacknife_pearson,
                }